# Copyright (C) 2017-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import logging
from typing import List

from swh.indexer.codemeta import SCHEMA_URI, compact, merge_values


class BaseMapping:
    """Base class for mappings to inherit from

    To implement a new mapping:

    - inherit this class
    - override translate function
    """

    def __init__(self, log_suffix=""):
        self.log_suffix = log_suffix
        self.log = logging.getLogger(
            "%s.%s" % (self.__class__.__module__, self.__class__.__name__)
        )

    @property
    def name(self):
        """A name of this mapping, used as an identifier in the
        indexer storage."""
        raise NotImplementedError(f"{self.__class__.__name__}.name")

    @classmethod
    def detect_metadata_files(cls, files):
        """
        Detects files potentially containing metadata

        Args:
            file_entries (list): list of files

        Returns:
            list: list of sha1 (possibly empty)
        """
        raise NotImplementedError(f"{cls.__name__}.detect_metadata_files")

    def translate(self, file_content):
        raise NotImplementedError(f"{self.__class__.__name__}.translate")

    def normalize_translation(self, metadata):
        return compact(metadata)


class SingleFileMapping(BaseMapping):
    """Base class for all mappings that use a single file as input."""

    @property
    def filename(self):
        """The .json file to extract metadata from."""
        raise NotImplementedError(f"{self.__class__.__name__}.filename")

    @classmethod
    def detect_metadata_files(cls, file_entries):
        for entry in file_entries:
            if entry["name"] == cls.filename:
                return [entry["sha1"]]
        return []


class DictMapping(BaseMapping):
    """Base class for mappings that take as input a file that is mostly
    a key-value store (eg. a shallow JSON dict)."""

    string_fields = []  # type: List[str]
    """List of fields that are simple strings, and don't need any
    normalization."""

    @property
    def mapping(self):
        """A translation dict to map dict keys into a canonical name."""
        raise NotImplementedError(f"{self.__class__.__name__}.mapping")

    @staticmethod
    def _normalize_method_name(name):
        return name.replace("-", "_")

    @classmethod
    def supported_terms(cls):
        return {
            term
            for (key, term) in cls.mapping.items()
            if key in cls.string_fields
            or hasattr(cls, "translate_" + cls._normalize_method_name(key))
            or hasattr(cls, "normalize_" + cls._normalize_method_name(key))
        }

    def _translate_dict(self, content_dict, *, normalize=True):
        """
        Translates content  by parsing content from a dict object
        and translating with the appropriate mapping

        Args:
            content_dict (dict): content dict to translate

        Returns:
            dict: translated metadata in json-friendly form needed for
            the indexer

        """
        translated_metadata = {"@type": SCHEMA_URI + "SoftwareSourceCode"}
        for k, v in content_dict.items():
            # First, check if there is a specific translation
            # method for this key
            translation_method = getattr(
                self, "translate_" + self._normalize_method_name(k), None
            )
            if translation_method:
                translation_method(translated_metadata, v)
            elif k in self.mapping:
                # if there is no method, but the key is known from the
                # crosswalk table
                codemeta_key = self.mapping[k]

                # if there is a normalization method, use it on the value
                normalization_method = getattr(
                    self, "normalize_" + self._normalize_method_name(k), None
                )
                if normalization_method:
                    v = normalization_method(v)
                elif k in self.string_fields and isinstance(v, str):
                    pass
                elif k in self.string_fields and isinstance(v, list):
                    v = [x for x in v if isinstance(x, str)]
                else:
                    continue

                # set the translation metadata with the normalized value
                if codemeta_key in translated_metadata:
                    translated_metadata[codemeta_key] = merge_values(
                        translated_metadata[codemeta_key], v
                    )
                else:
                    translated_metadata[codemeta_key] = v
        if normalize:
            return self.normalize_translation(translated_metadata)
        else:
            return translated_metadata


class JsonMapping(DictMapping, SingleFileMapping):
    """Base class for all mappings that use a JSON file as input."""

    def translate(self, raw_content):
        """
        Translates content by parsing content from a bytestring containing
        json data and translating with the appropriate mapping

        Args:
            raw_content (bytes): raw content to translate

        Returns:
            dict: translated metadata in json-friendly form needed for
            the indexer

        """
        try:
            raw_content = raw_content.decode()
        except UnicodeDecodeError:
            self.log.warning("Error unidecoding from %s", self.log_suffix)
            return
        try:
            content_dict = json.loads(raw_content)
        except json.JSONDecodeError:
            self.log.warning("Error unjsoning from %s", self.log_suffix)
            return
        if isinstance(content_dict, dict):
            return self._translate_dict(content_dict)



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import email.parser
import email.policy
import itertools

from swh.indexer.codemeta import CROSSWALK_TABLE, SCHEMA_URI

from .base import DictMapping, SingleFileMapping

_normalize_pkginfo_key = str.lower


class LinebreakPreservingEmailPolicy(email.policy.EmailPolicy):
    def header_fetch_parse(self, name, value):
        if hasattr(value, "name"):
            return value
        value = value.replace("\n        ", "\n")
        return self.header_factory(name, value)


class PythonPkginfoMapping(DictMapping, SingleFileMapping):
    """Dedicated class for Python's PKG-INFO mapping and translation.

    https://www.python.org/dev/peps/pep-0314/"""

    name = "pkg-info"
    filename = b"PKG-INFO"
    mapping = {
        _normalize_pkginfo_key(k): v
        for (k, v) in CROSSWALK_TABLE["Python PKG-INFO"].items()
    }
    string_fields = [
        "name",
        "version",
        "description",
        "summary",
        "author",
        "author-email",
    ]

    _parser = email.parser.BytesHeaderParser(policy=LinebreakPreservingEmailPolicy())

    def translate(self, content):
        msg = self._parser.parsebytes(content)
        d = {}
        for (key, value) in msg.items():
            key = _normalize_pkginfo_key(key)
            if value != "UNKNOWN":
                d.setdefault(key, []).append(value)
        metadata = self._translate_dict(d, normalize=False)
        if SCHEMA_URI + "author" in metadata or SCHEMA_URI + "email" in metadata:
            metadata[SCHEMA_URI + "author"] = {
                "@list": [
                    {
                        "@type": SCHEMA_URI + "Person",
                        SCHEMA_URI
                        + "name": metadata.pop(SCHEMA_URI + "author", [None])[0],
                        SCHEMA_URI
                        + "email": metadata.pop(SCHEMA_URI + "email", [None])[0],
                    }
                ]
            }
        return self.normalize_translation(metadata)

    def normalize_home_page(self, urls):
        return [{"@id": url} for url in urls]

    def normalize_keywords(self, keywords):
        return list(itertools.chain.from_iterable(s.split(" ") for s in keywords))

    def normalize_license(self, licenses):
        return [{"@id": license} for license in licenses]



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import unittest
from unittest.mock import patch

import pytest

import swh.indexer.ctags
from swh.indexer.ctags import CtagsIndexer, run_ctags
from swh.indexer.storage.model import ContentCtagsRow
from swh.indexer.tests.utils import (
    BASE_TEST_CONFIG,
    OBJ_STORAGE_DATA,
    SHA1_TO_CTAGS,
    CommonContentIndexerTest,
    fill_obj_storage,
    fill_storage,
    filter_dict,
)
from swh.model.hashutil import hash_to_bytes


class BasicTest(unittest.TestCase):
    @patch("swh.indexer.ctags.subprocess")
    def test_run_ctags(self, mock_subprocess):
        """Computing licenses from a raw content should return results

        """
        output0 = """
{"name":"defun","kind":"function","line":1,"language":"scheme"}
{"name":"name","kind":"symbol","line":5,"language":"else"}"""
        output1 = """
{"name":"let","kind":"var","line":10,"language":"something"}"""

        expected_result0 = [
            {"name": "defun", "kind": "function", "line": 1, "lang": "scheme"},
            {"name": "name", "kind": "symbol", "line": 5, "lang": "else"},
        ]

        expected_result1 = [
            {"name": "let", "kind": "var", "line": 10, "lang": "something"}
        ]
        for path, lang, intermediary_result, expected_result in [
            (b"some/path", "lisp", output0, expected_result0),
            (b"some/path/2", "markdown", output1, expected_result1),
        ]:
            mock_subprocess.check_output.return_value = intermediary_result
            actual_result = list(run_ctags(path, lang=lang))
            self.assertEqual(actual_result, expected_result)


class InjectCtagsIndexer:
    """Override ctags computations.

    """

    def compute_ctags(self, path, lang):
        """Inject fake ctags given path (sha1 identifier).

        """
        return {"lang": lang, **SHA1_TO_CTAGS.get(path)}


CONFIG = {
    **BASE_TEST_CONFIG,
    "tools": {
        "name": "universal-ctags",
        "version": "~git7859817b",
        "configuration": {
            "command_line": """ctags --fields=+lnz --sort=no """
            """ --links=no <filepath>""",
            "max_content_size": 1000,
        },
    },
    "languages": {"python": "python", "haskell": "haskell", "bar": "bar",},
    "workdir": "/tmp",
}


class TestCtagsIndexer(CommonContentIndexerTest, unittest.TestCase):
    """Ctags indexer test scenarios:

    - Known sha1s in the input list have their data indexed
    - Unknown sha1 in the input list are not indexed

    """

    def get_indexer_results(self, ids):
        yield from self.idx_storage.content_ctags_get(ids)

    def setUp(self):
        super().setUp()
        self.indexer = CtagsIndexer(config=CONFIG)
        self.indexer.catch_exceptions = False
        self.idx_storage = self.indexer.idx_storage
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        # Prepare test input
        self.id0 = "01c9379dfc33803963d07c1ccc748d3fe4c96bb5"
        self.id1 = "d4c647f0fc257591cc9ba1722484229780d1c607"
        self.id2 = "688a5ef812c53907562fe379d4b3851e69c7cb15"

        tool = {k.replace("tool_", ""): v for (k, v) in self.indexer.tool.items()}

        self.expected_results = [
            *[
                ContentCtagsRow(id=hash_to_bytes(self.id0), tool=tool, **kwargs,)
                for kwargs in SHA1_TO_CTAGS[self.id0]
            ],
            *[
                ContentCtagsRow(id=hash_to_bytes(self.id1), tool=tool, **kwargs,)
                for kwargs in SHA1_TO_CTAGS[self.id1]
            ],
            *[
                ContentCtagsRow(id=hash_to_bytes(self.id2), tool=tool, **kwargs,)
                for kwargs in SHA1_TO_CTAGS[self.id2]
            ],
        ]

        self._set_mocks()

    def _set_mocks(self):
        def find_ctags_for_content(raw_content):
            for (sha1, ctags) in SHA1_TO_CTAGS.items():
                if OBJ_STORAGE_DATA[sha1] == raw_content:
                    return ctags
            else:
                raise ValueError(
                    ("%r not found in objstorage, can't mock its ctags.") % raw_content
                )

        def fake_language(raw_content, *args, **kwargs):
            ctags = find_ctags_for_content(raw_content)
            return {"lang": ctags[0]["lang"]}

        self._real_compute_language = swh.indexer.ctags.compute_language
        swh.indexer.ctags.compute_language = fake_language

        def fake_check_output(cmd, *args, **kwargs):
            id_ = cmd[-1].split("/")[-1]
            return "\n".join(
                json.dumps({"language": ctag["lang"], **ctag})
                for ctag in SHA1_TO_CTAGS[id_]
            )

        self._real_check_output = swh.indexer.ctags.subprocess.check_output
        swh.indexer.ctags.subprocess.check_output = fake_check_output

    def tearDown(self):
        swh.indexer.ctags.compute_language = self._real_compute_language
        swh.indexer.ctags.subprocess.check_output = self._real_check_output
        super().tearDown()


def test_ctags_w_no_tool():
    with pytest.raises(ValueError):
        CtagsIndexer(config=filter_dict(CONFIG, "tools"))



# Copyright (C) 2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging

from swh.core.utils import grouper
from swh.scheduler.utils import create_task_dict

MAX_ORIGINS_PER_TASK = 100


def process_journal_objects(messages, *, scheduler, task_names):
    """Worker function for `JournalClient.process(worker_fn)`, after
    currification of `scheduler` and `task_names`."""
    assert set(messages) == {"origin_visit"}, set(messages)
    process_origin_visits(messages["origin_visit"], scheduler, task_names)


def process_origin_visits(visits, scheduler, task_names):
    task_dicts = []
    logging.debug("processing origin visits %r", visits)
    if task_names.get("origin_metadata"):
        visits = [visit for visit in visits if visit["status"] == "full"]
        visit_batches = grouper(visits, MAX_ORIGINS_PER_TASK)
        for visit_batch in visit_batches:
            visit_urls = []
            for visit in visit_batch:
                if isinstance(visit["origin"], str):
                    visit_urls.append(visit["origin"])
                else:
                    visit_urls.append(visit["origin"]["url"])
            task_dicts.append(
                create_task_dict(
                    task_names["origin_metadata"],
                    "oneshot",
                    visit_urls,
                    policy_update="update-dups",
                )
            )

    if task_dicts:
        scheduler.create_tasks(task_dicts)



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import os

from swh.core import config
from swh.core.api import RPCServerApp
from swh.core.api import encode_data_server as encode_data
from swh.core.api import error_handler
from swh.indexer.storage import INDEXER_CFG_KEY, get_indexer_storage
from swh.indexer.storage.exc import IndexerStorageArgumentException
from swh.indexer.storage.interface import IndexerStorageInterface

from .serializers import DECODERS, ENCODERS


def get_storage():
    global storage
    if not storage:
        storage = get_indexer_storage(**app.config[INDEXER_CFG_KEY])

    return storage


class IndexerStorageServerApp(RPCServerApp):
    extra_type_decoders = DECODERS
    extra_type_encoders = ENCODERS


app = IndexerStorageServerApp(
    __name__, backend_class=IndexerStorageInterface, backend_factory=get_storage
)
storage = None


@app.errorhandler(Exception)
def my_error_handler(exception):
    return error_handler(exception, encode_data)


@app.errorhandler(IndexerStorageArgumentException)
def argument_error_handler(exception):
    return error_handler(exception, encode_data, status_code=400)


@app.route("/")
def index():
    return "SWH Indexer Storage API server"


api_cfg = None


def load_and_check_config(config_file, type="local"):
    """Check the minimal configuration is set to run the api or raise an
       error explanation.

    Args:
        config_file (str): Path to the configuration file to load
        type (str): configuration type. For 'local' type, more
                    checks are done.

    Raises:
        Error if the setup is not as expected

    Returns:
        configuration as a dict

    """
    if not config_file:
        raise EnvironmentError("Configuration file must be defined")

    if not os.path.exists(config_file):
        raise FileNotFoundError("Configuration file %s does not exist" % (config_file,))

    cfg = config.read(config_file)
    if "indexer_storage" not in cfg:
        raise KeyError("Missing '%indexer_storage' configuration")

    if type == "local":
        vcfg = cfg["indexer_storage"]
        cls = vcfg.get("cls")
        if cls != "local":
            raise ValueError(
                "The indexer_storage backend can only be started with a "
                "'local' configuration"
            )

        args = vcfg["args"]
        if not args.get("db"):
            raise ValueError("Invalid configuration; missing 'db' config entry")

    return cfg


def make_app_from_configfile():
    """Run the WSGI app from the webserver, loading the configuration from
       a configuration file.

       SWH_CONFIG_FILENAME environment variable defines the
       configuration path to load.

    """
    global api_cfg
    if not api_cfg:
        config_file = os.environ.get("SWH_CONFIG_FILENAME")
        api_cfg = load_and_check_config(config_file)
        app.config.update(api_cfg)
    handler = logging.StreamHandler()
    app.logger.addHandler(handler)
    return app


if __name__ == "__main__":
    print("Deprecated. Use swh-indexer")



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from collections import defaultdict
import itertools
import logging
from typing import Any, Dict, Generator, List, Optional, Tuple

from swh.core import utils
from swh.core.config import load_from_envvar
from swh.model import hashutil
from swh.model.model import Content
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.factory import get_objstorage
from swh.storage import get_storage

DEFAULT_CONFIG: Dict[str, Any] = {
    "storage": {"cls": "memory"},
    "objstorage": {"cls": "memory"},
    # the set of checksums that should be computed.
    # Examples: 'sha1_git', 'blake2b512', 'blake2s256'
    "compute_checksums": [],
    # whether checksums that already exist in the DB should be
    # recomputed/updated or left untouched
    "recompute_checksums": False,
    # Number of contents to retrieve blobs at the same time
    "batch_size_retrieve_content": 10,
    # Number of contents to update at the same time
    "batch_size_update": 100,
}


class RecomputeChecksums:
    """Class in charge of (re)computing content's hashes.

    Hashes to compute are defined across 2 configuration options:

    compute_checksums ([str])
      list of hash algorithms that
      py:func:`swh.model.hashutil.MultiHash.from_data` function should
      be able to deal with. For variable-length checksums, a desired
      checksum length should also be provided. Their format is
      <algorithm's name>:<variable-length> e.g: blake2:512

    recompute_checksums (bool)
      a boolean to notify that we also want to recompute potential existing
      hashes specified in compute_checksums.  Default to False.

    """

    def __init__(self) -> None:
        self.config = load_from_envvar(DEFAULT_CONFIG)
        self.storage = get_storage(**self.config["storage"])
        self.objstorage = get_objstorage(**self.config["objstorage"])
        self.compute_checksums = self.config["compute_checksums"]
        self.recompute_checksums = self.config["recompute_checksums"]
        self.batch_size_retrieve_content = self.config["batch_size_retrieve_content"]
        self.batch_size_update = self.config["batch_size_update"]
        self.log = logging.getLogger("swh.indexer.rehash")

        if not self.compute_checksums:
            raise ValueError("Checksums list should not be empty.")

    def _read_content_ids(
        self, contents: List[Dict[str, Any]]
    ) -> Generator[bytes, Any, None]:
        """Read the content identifiers from the contents.

        """
        for c in contents:
            h = c["sha1"]
            if isinstance(h, str):
                h = hashutil.hash_to_bytes(h)

            yield h

    def get_new_contents_metadata(
        self, all_contents: List[Dict[str, Any]]
    ) -> Generator[Tuple[Dict[str, Any], List[Any]], Any, None]:
        """Retrieve raw contents and compute new checksums on the
           contents. Unknown or corrupted contents are skipped.

        Args:
            all_contents: List of contents as dictionary with
              the necessary primary keys

        Yields:
            tuple: tuple of (content to update, list of checksums computed)

        """
        content_ids = self._read_content_ids(all_contents)
        for contents in utils.grouper(content_ids, self.batch_size_retrieve_content):
            contents_iter = itertools.tee(contents, 2)
            try:
                sha1s = [s for s in contents_iter[0]]
                content_metadata: List[Optional[Content]] = self.storage.content_get(
                    sha1s
                )
            except Exception:
                self.log.exception("Problem when reading contents metadata.")
                continue

            for sha1, content_model in zip(sha1s, content_metadata):
                if not content_model:
                    continue
                content: Dict = content_model.to_dict()
                # Recompute checksums provided in compute_checksums options
                if self.recompute_checksums:
                    checksums_to_compute = list(self.compute_checksums)
                else:
                    # Compute checksums provided in compute_checksums
                    # options not already defined for that content
                    checksums_to_compute = [
                        h for h in self.compute_checksums if not content.get(h)
                    ]

                if not checksums_to_compute:  # Nothing to recompute
                    continue

                try:
                    raw_content = self.objstorage.get(sha1)
                except ObjNotFoundError:
                    self.log.warning("Content %s not found in objstorage!", sha1)
                    continue

                content_hashes = hashutil.MultiHash.from_data(
                    raw_content, hash_names=checksums_to_compute
                ).digest()
                content.update(content_hashes)
                yield content, checksums_to_compute

    def run(self, contents: List[Dict[str, Any]]) -> Dict:
        """Given a list of content:

          - (re)compute a given set of checksums on contents available in our
            object storage
          - update those contents with the new metadata

        Args:
            contents: contents as dictionary with necessary keys.
                key present in such dictionary should be the ones defined in
                the 'primary_key' option.

        Returns:
            A summary dict with key 'status', task' status and 'count' the
            number of updated contents.

        """
        status = "uneventful"
        count = 0
        for data in utils.grouper(
            self.get_new_contents_metadata(contents), self.batch_size_update
        ):

            groups: Dict[str, List[Any]] = defaultdict(list)
            for content, keys_to_update in data:
                keys_str = ",".join(keys_to_update)
                groups[keys_str].append(content)

            for keys_to_update, contents in groups.items():
                keys: List[str] = keys_to_update.split(",")
                try:
                    self.storage.content_update(contents, keys=keys)
                    count += len(contents)
                    status = "eventful"
                except Exception:
                    self.log.exception("Problem during update.")
                    continue

        return {
            "status": status,
            "count": count,
        }



from os import path

import celery.contrib.testing.tasks  # noqa
from celery.contrib.testing.worker import start_worker

import swh.indexer
from swh.scheduler.celery_backend.config import app

__all__ = ["start_worker_thread"]

SQL_DIR = path.join(path.dirname(swh.indexer.__file__), "sql")


def start_worker_thread():
    return start_worker(app)



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
import re
from typing import Any, Dict, List, Tuple, Union

import click

from swh.indexer.indexer import OriginIndexer
from swh.model.model import SnapshotBranch, TargetType
from swh.storage.algos.origin import origin_get_latest_visit_status
from swh.storage.algos.snapshot import snapshot_get_all_branches


class OriginHeadIndexer(OriginIndexer[Dict]):
    """Origin-level indexer.

    This indexer is in charge of looking up the revision that acts as the
    "head" of an origin.

    In git, this is usually the commit pointed to by the 'master' branch."""

    USE_TOOLS = False

    def persist_index_computations(
        self, results: Any, policy_update: str
    ) -> Dict[str, int]:
        """Do nothing. The indexer's results are not persistent, they
        should only be piped to another indexer."""
        return {}

    # Dispatch

    def index(self, id: str, data: None = None, **kwargs) -> List[Dict]:
        origin_url = id
        visit_and_status = origin_get_latest_visit_status(
            self.storage, origin_url, allowed_statuses=["full"], require_snapshot=True
        )
        if not visit_and_status:
            return []
        visit, visit_status = visit_and_status
        assert visit_status.snapshot is not None
        snapshot = snapshot_get_all_branches(self.storage, visit_status.snapshot)
        if snapshot is None:
            return []
        method = getattr(
            self, "_try_get_%s_head" % visit.type, self._try_get_head_generic
        )

        rev_id = method(snapshot.branches)
        if rev_id is not None:
            return [{"origin_url": origin_url, "revision_id": rev_id,}]

        # could not find a head revision
        return []

    # Tarballs

    _archive_filename_re = re.compile(
        rb"^"
        rb"(?P<pkgname>.*)[-_]"
        rb"(?P<version>[0-9]+(\.[0-9])*)"
        rb"(?P<preversion>[-+][a-zA-Z0-9.~]+?)?"
        rb"(?P<extension>(\.[a-zA-Z0-9]+)+)"
        rb"$"
    )

    @classmethod
    def _parse_version(cls: Any, filename: bytes) -> Tuple[Union[float, int], ...]:
        """Extracts the release version from an archive filename,
        to get an ordering whose maximum is likely to be the last
        version of the software

        >>> OriginHeadIndexer._parse_version(b'foo')
        (-inf,)
        >>> OriginHeadIndexer._parse_version(b'foo.tar.gz')
        (-inf,)
        >>> OriginHeadIndexer._parse_version(b'gnu-hello-0.0.1.tar.gz')
        (0, 0, 1, 0)
        >>> OriginHeadIndexer._parse_version(b'gnu-hello-0.0.1-beta2.tar.gz')
        (0, 0, 1, -1, 'beta2')
        >>> OriginHeadIndexer._parse_version(b'gnu-hello-0.0.1+foobar.tar.gz')
        (0, 0, 1, 1, 'foobar')
        """
        res = cls._archive_filename_re.match(filename)
        if res is None:
            return (float("-infinity"),)
        version = [int(n) for n in res.group("version").decode().split(".")]
        if res.group("preversion") is None:
            version.append(0)
        else:
            preversion = res.group("preversion").decode()
            if preversion.startswith("-"):
                version.append(-1)
                version.append(preversion[1:])
            elif preversion.startswith("+"):
                version.append(1)
                version.append(preversion[1:])
            else:
                assert False, res.group("preversion")
        return tuple(version)

    def _try_get_ftp_head(self, branches: Dict[bytes, SnapshotBranch]) -> Any:
        archive_names = list(branches)
        max_archive_name = max(archive_names, key=self._parse_version)
        r = self._try_resolve_target(branches, max_archive_name)
        return r

    # Generic

    def _try_get_head_generic(self, branches: Dict[bytes, SnapshotBranch]) -> Any:
        # Works on 'deposit', 'pypi', and VCSs.
        return self._try_resolve_target(branches, b"HEAD") or self._try_resolve_target(
            branches, b"master"
        )

    def _try_resolve_target(
        self, branches: Dict[bytes, SnapshotBranch], branch_name: bytes
    ) -> Any:
        try:
            branch = branches[branch_name]
            if branch is None:
                return None
            while branch.target_type == TargetType.ALIAS:
                branch = branches[branch.target]
                if branch is None:
                    return None

            if branch.target_type == TargetType.REVISION:
                return branch.target
            elif branch.target_type == TargetType.CONTENT:
                return None  # TODO
            elif branch.target_type == TargetType.DIRECTORY:
                return None  # TODO
            elif branch.target_type == TargetType.RELEASE:
                return None  # TODO
            else:
                assert False, branch
        except KeyError:
            return None


@click.command()
@click.option(
    "--origins", "-i", help='Origins to lookup, in the "type+url" format', multiple=True
)
def main(origins: List[str]) -> None:
    rev_metadata_indexer = OriginHeadIndexer()
    rev_metadata_indexer.run(origins)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    main()



# Copyright (C) 2018-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from unittest.mock import patch

from swh.indexer.metadata import OriginMetadataIndexer
from swh.indexer.storage.interface import IndexerStorageInterface
from swh.indexer.storage.model import (
    OriginIntrinsicMetadataRow,
    RevisionIntrinsicMetadataRow,
)
from swh.model.model import Origin
from swh.storage.interface import StorageInterface

from .test_metadata import REVISION_METADATA_CONFIG
from .utils import REVISION, YARN_PARSER_METADATA


def test_origin_metadata_indexer(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    origin = "https://github.com/librariesio/yarn-parser"
    indexer.run([origin])

    tool = {
        "name": "swh-metadata-translator",
        "version": "0.0.2",
        "configuration": {"context": "NpmMapping", "type": "local"},
    }

    rev_id = REVISION.id
    rev_metadata = RevisionIntrinsicMetadataRow(
        id=rev_id, tool=tool, metadata=YARN_PARSER_METADATA, mappings=["npm"],
    )
    origin_metadata = OriginIntrinsicMetadataRow(
        id=origin,
        tool=tool,
        from_revision=rev_id,
        metadata=YARN_PARSER_METADATA,
        mappings=["npm"],
    )

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    for rev_result in rev_results:
        assert rev_result.tool
        del rev_result.tool["id"]
    assert rev_results == [rev_metadata]

    orig_results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    for orig_result in orig_results:
        assert orig_result.tool
        del orig_result.tool["id"]
    assert orig_results == [origin_metadata]


def test_origin_metadata_indexer_duplicate_origin(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:
    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.storage = storage
    indexer.idx_storage = idx_storage
    indexer.run(["https://github.com/librariesio/yarn-parser"])
    indexer.run(["https://github.com/librariesio/yarn-parser"] * 2)

    origin = "https://github.com/librariesio/yarn-parser"
    rev_id = REVISION.id

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert len(rev_results) == 1

    orig_results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    assert len(orig_results) == 1


def test_origin_metadata_indexer_missing_head(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:
    storage.origin_add([Origin(url="https://example.com")])

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.run(["https://example.com"])

    origin = "https://example.com"

    results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    assert results == []


def test_origin_metadata_indexer_partial_missing_head(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:

    origin1 = "https://example.com"
    origin2 = "https://github.com/librariesio/yarn-parser"
    storage.origin_add([Origin(url=origin1)])
    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.run([origin1, origin2])

    rev_id = REVISION.id

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert rev_results == [
        RevisionIntrinsicMetadataRow(
            id=rev_id,
            metadata=YARN_PARSER_METADATA,
            mappings=["npm"],
            tool=rev_results[0].tool,
        )
    ]

    orig_results = list(
        indexer.idx_storage.origin_intrinsic_metadata_get([origin1, origin2])
    )
    for orig_result in orig_results:
        assert orig_results == [
            OriginIntrinsicMetadataRow(
                id=origin2,
                from_revision=rev_id,
                metadata=YARN_PARSER_METADATA,
                mappings=["npm"],
                tool=orig_results[0].tool,
            )
        ]


def test_origin_metadata_indexer_duplicate_revision(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:
    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    indexer.storage = storage
    indexer.idx_storage = idx_storage
    indexer.catch_exceptions = False
    origin1 = "https://github.com/librariesio/yarn-parser"
    origin2 = "https://github.com/librariesio/yarn-parser.git"
    indexer.run([origin1, origin2])

    rev_id = REVISION.id

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert len(rev_results) == 1

    orig_results = list(
        indexer.idx_storage.origin_intrinsic_metadata_get([origin1, origin2])
    )
    assert len(orig_results) == 2


def test_origin_metadata_indexer_no_metadata_file(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    origin = "https://github.com/librariesio/yarn-parser"
    with patch("swh.indexer.metadata_dictionary.npm.NpmMapping.filename", b"foo.json"):
        indexer.run([origin])

    rev_id = REVISION.id

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert rev_results == []

    orig_results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    assert orig_results == []


def test_origin_metadata_indexer_no_metadata(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    origin = "https://github.com/librariesio/yarn-parser"
    with patch(
        "swh.indexer.metadata.RevisionMetadataIndexer"
        ".translate_revision_intrinsic_metadata",
        return_value=(["npm"], {"@context": "foo"}),
    ):
        indexer.run([origin])

    rev_id = REVISION.id

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert rev_results == []

    orig_results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    assert orig_results == []


def test_origin_metadata_indexer_error(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    origin = "https://github.com/librariesio/yarn-parser"
    with patch(
        "swh.indexer.metadata.RevisionMetadataIndexer"
        ".translate_revision_intrinsic_metadata",
        return_value=None,
    ):
        indexer.run([origin])

    rev_id = REVISION.id

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert rev_results == []

    orig_results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    assert orig_results == []


def test_origin_metadata_indexer_delete_metadata(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    origin = "https://github.com/librariesio/yarn-parser"
    indexer.run([origin])

    rev_id = REVISION.id

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert rev_results != []

    orig_results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    assert orig_results != []

    with patch("swh.indexer.metadata_dictionary.npm.NpmMapping.filename", b"foo.json"):
        indexer.run([origin])

    rev_results = list(indexer.idx_storage.revision_intrinsic_metadata_get([rev_id]))
    assert rev_results == []

    orig_results = list(indexer.idx_storage.origin_intrinsic_metadata_get([origin]))
    assert orig_results == []


def test_origin_metadata_indexer_unknown_origin(
    idx_storage: IndexerStorageInterface, storage: StorageInterface, obj_storage
) -> None:

    indexer = OriginMetadataIndexer(config=REVISION_METADATA_CONFIG)
    result = indexer.index_list(["https://unknown.org/foo"])
    assert not result



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


def test_task_origin_metadata(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):

    mock_indexer = mocker.patch("swh.indexer.tasks.OriginMetadataIndexer.run")
    mock_indexer.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.indexer.tasks.OriginMetadata", args=["origin-url"],
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}


def test_task_ctags(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):

    mock_indexer = mocker.patch("swh.indexer.tasks.CtagsIndexer.run")
    mock_indexer.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task("swh.indexer.tasks.Ctags", args=["id0"],)
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}


def test_task_fossology_license(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):

    mock_indexer = mocker.patch("swh.indexer.tasks.FossologyLicenseIndexer.run")
    mock_indexer.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.indexer.tasks.ContentFossologyLicense", args=["id0"],
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}


def test_task_recompute_checksums(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):

    mock_indexer = mocker.patch("swh.indexer.tasks.RecomputeChecksums.run")
    mock_indexer.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.indexer.tasks.RecomputeChecksums", args=[[{"blake2b256": "id"}]],
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}


def test_task_mimetype(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):

    mock_indexer = mocker.patch("swh.indexer.tasks.MimetypeIndexer.run")
    mock_indexer.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.indexer.tasks.ContentMimetype", args=["id0"],
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}


def test_task_mimetype_partition(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):

    mock_indexer = mocker.patch("swh.indexer.tasks.MimetypePartitionIndexer.run")
    mock_indexer.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.indexer.tasks.ContentMimetypePartition", args=[0, 4],
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}


def test_task_license_partition(
    mocker, swh_scheduler_celery_app, swh_scheduler_celery_worker, swh_config
):

    mock_indexer = mocker.patch(
        "swh.indexer.tasks.FossologyLicensePartitionIndexer.run"
    )
    mock_indexer.return_value = {"status": "eventful"}

    res = swh_scheduler_celery_app.send_task(
        "swh.indexer.tasks.ContentFossologyLicensePartition", args=[0, 4],
    )
    assert res
    res.wait()
    assert res.successful()

    assert res.result == {"status": "eventful"}



# Copyright (C) 2017-2018  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict
import unittest
from unittest.mock import patch

import pytest

from swh.indexer import fossology_license
from swh.indexer.fossology_license import (
    FossologyLicenseIndexer,
    FossologyLicensePartitionIndexer,
    compute_license,
)
from swh.indexer.storage.model import ContentLicenseRow
from swh.indexer.tests.utils import (
    BASE_TEST_CONFIG,
    SHA1_TO_LICENSES,
    CommonContentIndexerPartitionTest,
    CommonContentIndexerTest,
    fill_obj_storage,
    fill_storage,
    filter_dict,
)
from swh.model.hashutil import hash_to_bytes


class BasicTest(unittest.TestCase):
    @patch("swh.indexer.fossology_license.subprocess")
    def test_compute_license(self, mock_subprocess):
        """Computing licenses from a raw content should return results

        """
        for path, intermediary_result, output in [
            (b"some/path", None, []),
            (b"some/path/2", [], []),
            (b"other/path", " contains license(s) GPL,AGPL", ["GPL", "AGPL"]),
        ]:
            mock_subprocess.check_output.return_value = intermediary_result

            actual_result = compute_license(path)

            self.assertEqual(actual_result, {"licenses": output, "path": path,})


def mock_compute_license(path):
    """path is the content identifier

    """
    if isinstance(id, bytes):
        path = path.decode("utf-8")
    # path is something like /tmp/tmpXXX/<sha1> so we keep only the sha1 part
    path = path.split("/")[-1]
    return {"licenses": SHA1_TO_LICENSES.get(path, [])}


CONFIG = {
    **BASE_TEST_CONFIG,
    "workdir": "/tmp",
    "tools": {
        "name": "nomos",
        "version": "3.1.0rc2-31-ga2cbb8c",
        "configuration": {"command_line": "nomossa <filepath>",},
    },
}  # type: Dict[str, Any]

RANGE_CONFIG = dict(list(CONFIG.items()) + [("write_batch_size", 100)])


class TestFossologyLicenseIndexer(CommonContentIndexerTest, unittest.TestCase):
    """Language indexer test scenarios:

    - Known sha1s in the input list have their data indexed
    - Unknown sha1 in the input list are not indexed

    """

    def get_indexer_results(self, ids):
        yield from self.idx_storage.content_fossology_license_get(ids)

    def setUp(self):
        super().setUp()
        # replace actual license computation with a mock
        self.orig_compute_license = fossology_license.compute_license
        fossology_license.compute_license = mock_compute_license

        self.indexer = FossologyLicenseIndexer(CONFIG)
        self.indexer.catch_exceptions = False
        self.idx_storage = self.indexer.idx_storage
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

        self.id0 = "01c9379dfc33803963d07c1ccc748d3fe4c96bb5"
        self.id1 = "688a5ef812c53907562fe379d4b3851e69c7cb15"
        self.id2 = "da39a3ee5e6b4b0d3255bfef95601890afd80709"  # empty content

        tool = {k.replace("tool_", ""): v for (k, v) in self.indexer.tool.items()}
        # then
        self.expected_results = [
            *[
                ContentLicenseRow(
                    id=hash_to_bytes(self.id0), tool=tool, license=license
                )
                for license in SHA1_TO_LICENSES[self.id0]
            ],
            *[
                ContentLicenseRow(
                    id=hash_to_bytes(self.id1), tool=tool, license=license
                )
                for license in SHA1_TO_LICENSES[self.id1]
            ],
            *[],  # self.id2
        ]

    def tearDown(self):
        super().tearDown()
        fossology_license.compute_license = self.orig_compute_license


class TestFossologyLicensePartitionIndexer(
    CommonContentIndexerPartitionTest, unittest.TestCase
):
    """Range Fossology License Indexer tests.

    - new data within range are indexed
    - no data outside a range are indexed
    - with filtering existing indexed data prior to compute new index
    - without filtering existing indexed data prior to compute new index

    """

    def setUp(self):
        super().setUp()

        # replace actual license computation with a mock
        self.orig_compute_license = fossology_license.compute_license
        fossology_license.compute_license = mock_compute_license

        self.indexer = FossologyLicensePartitionIndexer(config=RANGE_CONFIG)
        self.indexer.catch_exceptions = False
        fill_storage(self.indexer.storage)
        fill_obj_storage(self.indexer.objstorage)

    def tearDown(self):
        super().tearDown()
        fossology_license.compute_license = self.orig_compute_license


def test_fossology_w_no_tool():
    with pytest.raises(ValueError):
        FossologyLicenseIndexer(config=filter_dict(CONFIG, "tools"))


def test_fossology_range_w_no_tool():
    with pytest.raises(ValueError):
        FossologyLicensePartitionIndexer(config=filter_dict(RANGE_CONFIG, "tools"))



# Copyright (C) 2017-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
import unittest

from hypothesis import HealthCheck, given, settings, strategies

from swh.indexer.codemeta import CODEMETA_TERMS
from swh.indexer.metadata import ContentMetadataIndexer, RevisionMetadataIndexer
from swh.indexer.metadata_detector import detect_metadata
from swh.indexer.metadata_dictionary import MAPPINGS
from swh.indexer.metadata_dictionary.maven import MavenMapping
from swh.indexer.metadata_dictionary.npm import NpmMapping
from swh.indexer.metadata_dictionary.ruby import GemspecMapping
from swh.indexer.storage.model import ContentMetadataRow, RevisionIntrinsicMetadataRow
from swh.indexer.tests.utils import DIRECTORY2, REVISION
from swh.model.hashutil import hash_to_bytes
from swh.model.model import Directory, DirectoryEntry, Revision

from .utils import (
    BASE_TEST_CONFIG,
    YARN_PARSER_METADATA,
    fill_obj_storage,
    fill_storage,
    json_document_strategy,
    xml_document_strategy,
)

TRANSLATOR_TOOL = {
    "name": "swh-metadata-translator",
    "version": "0.0.2",
    "configuration": {"type": "local", "context": "NpmMapping"},
}


class ContentMetadataTestIndexer(ContentMetadataIndexer):
    """Specific Metadata whose configuration is enough to satisfy the
       indexing tests.
    """

    def parse_config_file(self, *args, **kwargs):
        assert False, "should not be called; the rev indexer configures it."


REVISION_METADATA_CONFIG = {
    **BASE_TEST_CONFIG,
    "tools": TRANSLATOR_TOOL,
}


class Metadata(unittest.TestCase):
    """
    Tests metadata_mock_tool tool for Metadata detection
    """

    def setUp(self):
        """
        shows the entire diff in the results
        """
        self.maxDiff = None
        self.npm_mapping = MAPPINGS["NpmMapping"]()
        self.codemeta_mapping = MAPPINGS["CodemetaMapping"]()
        self.maven_mapping = MAPPINGS["MavenMapping"]()
        self.pkginfo_mapping = MAPPINGS["PythonPkginfoMapping"]()
        self.gemspec_mapping = MAPPINGS["GemspecMapping"]()

    def test_compute_metadata_none(self):
        """
        testing content empty content is empty
        should return None
        """
        # given
        content = b""

        # None if no metadata was found or an error occurred
        declared_metadata = None
        # when
        result = self.npm_mapping.translate(content)
        # then
        self.assertEqual(declared_metadata, result)

    def test_compute_metadata_npm(self):
        """
        testing only computation of metadata with hard_mapping_npm
        """
        # given
        content = b"""
            {
                "name": "test_metadata",
                "version": "0.0.2",
                "description": "Simple package.json test for indexer",
                  "repository": {
                    "type": "git",
                    "url": "https://github.com/moranegg/metadata_test"
                },
                "author": {
                    "email": "moranegg@example.com",
                    "name": "Morane G"
                }
            }
        """
        declared_metadata = {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "type": "SoftwareSourceCode",
            "name": "test_metadata",
            "version": "0.0.2",
            "description": "Simple package.json test for indexer",
            "codeRepository": "git+https://github.com/moranegg/metadata_test",
            "author": [
                {"type": "Person", "name": "Morane G", "email": "moranegg@example.com",}
            ],
        }

        # when
        result = self.npm_mapping.translate(content)
        # then
        self.assertEqual(declared_metadata, result)

    def test_index_content_metadata_npm(self):
        """
        testing NPM with package.json
        - one sha1 uses a file that can't be translated to metadata and
          should return None in the translated metadata
        """
        # given
        sha1s = [
            hash_to_bytes("26a9f72a7c87cc9205725cfd879f514ff4f3d8d5"),
            hash_to_bytes("d4c647f0fc257591cc9ba1722484229780d1c607"),
            hash_to_bytes("02fb2c89e14f7fab46701478c83779c7beb7b069"),
        ]
        # this metadata indexer computes only metadata for package.json
        # in npm context with a hard mapping
        config = BASE_TEST_CONFIG.copy()
        config["tools"] = [TRANSLATOR_TOOL]
        metadata_indexer = ContentMetadataTestIndexer(config=config)
        fill_obj_storage(metadata_indexer.objstorage)
        fill_storage(metadata_indexer.storage)

        # when
        metadata_indexer.run(sha1s, policy_update="ignore-dups")
        results = list(metadata_indexer.idx_storage.content_metadata_get(sha1s))

        expected_results = [
            ContentMetadataRow(
                id=hash_to_bytes("26a9f72a7c87cc9205725cfd879f514ff4f3d8d5"),
                tool=TRANSLATOR_TOOL,
                metadata={
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "type": "SoftwareSourceCode",
                    "codeRepository": "git+https://github.com/moranegg/metadata_test",
                    "description": "Simple package.json test for indexer",
                    "name": "test_metadata",
                    "version": "0.0.1",
                },
            ),
            ContentMetadataRow(
                id=hash_to_bytes("d4c647f0fc257591cc9ba1722484229780d1c607"),
                tool=TRANSLATOR_TOOL,
                metadata={
                    "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                    "type": "SoftwareSourceCode",
                    "issueTracker": "https://github.com/npm/npm/issues",
                    "author": [
                        {
                            "type": "Person",
                            "name": "Isaac Z. Schlueter",
                            "email": "i@izs.me",
                            "url": "http://blog.izs.me",
                        }
                    ],
                    "codeRepository": "git+https://github.com/npm/npm",
                    "description": "a package manager for JavaScript",
                    "license": "https://spdx.org/licenses/Artistic-2.0",
                    "version": "5.0.3",
                    "name": "npm",
                    "keywords": [
                        "install",
                        "modules",
                        "package manager",
                        "package.json",
                    ],
                    "url": "https://docs.npmjs.com/",
                },
            ),
        ]

        for result in results:
            del result.tool["id"]

        # The assertion below returns False sometimes because of nested lists
        self.assertEqual(expected_results, results)

    def test_npm_bugs_normalization(self):
        # valid dictionary
        package_json = b"""{
            "name": "foo",
            "bugs": {
                "url": "https://github.com/owner/project/issues",
                "email": "foo@example.com"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "name": "foo",
                "issueTracker": "https://github.com/owner/project/issues",
                "type": "SoftwareSourceCode",
            },
        )

        # "invalid" dictionary
        package_json = b"""{
            "name": "foo",
            "bugs": {
                "email": "foo@example.com"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "name": "foo",
                "type": "SoftwareSourceCode",
            },
        )

        # string
        package_json = b"""{
            "name": "foo",
            "bugs": "https://github.com/owner/project/issues"
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "name": "foo",
                "issueTracker": "https://github.com/owner/project/issues",
                "type": "SoftwareSourceCode",
            },
        )

    def test_npm_repository_normalization(self):
        # normal
        package_json = b"""{
            "name": "foo",
            "repository": {
                "type" : "git",
                "url" : "https://github.com/npm/cli.git"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "name": "foo",
                "codeRepository": "git+https://github.com/npm/cli.git",
                "type": "SoftwareSourceCode",
            },
        )

        # missing url
        package_json = b"""{
            "name": "foo",
            "repository": {
                "type" : "git"
            }
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "name": "foo",
                "type": "SoftwareSourceCode",
            },
        )

        # github shortcut
        package_json = b"""{
            "name": "foo",
            "repository": "github:npm/cli"
        }"""
        result = self.npm_mapping.translate(package_json)
        expected_result = {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "name": "foo",
            "codeRepository": "git+https://github.com/npm/cli.git",
            "type": "SoftwareSourceCode",
        }
        self.assertEqual(result, expected_result)

        # github shortshortcut
        package_json = b"""{
            "name": "foo",
            "repository": "npm/cli"
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(result, expected_result)

        # gitlab shortcut
        package_json = b"""{
            "name": "foo",
            "repository": "gitlab:user/repo"
        }"""
        result = self.npm_mapping.translate(package_json)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "name": "foo",
                "codeRepository": "git+https://gitlab.com/user/repo.git",
                "type": "SoftwareSourceCode",
            },
        )

    def test_detect_metadata_package_json(self):
        # given
        df = [
            {
                "sha1_git": b"abc",
                "name": b"index.js",
                "target": b"abc",
                "length": 897,
                "status": "visible",
                "type": "file",
                "perms": 33188,
                "dir_id": b"dir_a",
                "sha1": b"bcd",
            },
            {
                "sha1_git": b"aab",
                "name": b"package.json",
                "target": b"aab",
                "length": 712,
                "status": "visible",
                "type": "file",
                "perms": 33188,
                "dir_id": b"dir_a",
                "sha1": b"cde",
            },
        ]
        # when
        results = detect_metadata(df)

        expected_results = {"NpmMapping": [b"cde"]}
        # then
        self.assertEqual(expected_results, results)

    def test_compute_metadata_valid_codemeta(self):
        raw_content = b"""{
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "@type": "SoftwareSourceCode",
            "identifier": "CodeMeta",
            "description": "CodeMeta is a concept vocabulary that can be used to standardize the exchange of software metadata across repositories and organizations.",
            "name": "CodeMeta: Minimal metadata schemas for science software and code, in JSON-LD",
            "codeRepository": "https://github.com/codemeta/codemeta",
            "issueTracker": "https://github.com/codemeta/codemeta/issues",
            "license": "https://spdx.org/licenses/Apache-2.0",
            "version": "2.0",
            "author": [
              {
                "@type": "Person",
                "givenName": "Carl",
                "familyName": "Boettiger",
                "email": "cboettig@gmail.com",
                "@id": "http://orcid.org/0000-0002-1642-628X"
              },
              {
                "@type": "Person",
                "givenName": "Matthew B.",
                "familyName": "Jones",
                "email": "jones@nceas.ucsb.edu",
                "@id": "http://orcid.org/0000-0003-0077-4738"
              }
            ],
            "maintainer": {
              "@type": "Person",
              "givenName": "Carl",
              "familyName": "Boettiger",
              "email": "cboettig@gmail.com",
              "@id": "http://orcid.org/0000-0002-1642-628X"
            },
            "contIntegration": "https://travis-ci.org/codemeta/codemeta",
            "developmentStatus": "active",
            "downloadUrl": "https://github.com/codemeta/codemeta/archive/2.0.zip",
            "funder": {
                "@id": "https://doi.org/10.13039/100000001",
                "@type": "Organization",
                "name": "National Science Foundation"
            },
            "funding":"1549758; Codemeta: A Rosetta Stone for Metadata in Scientific Software",
            "keywords": [
              "metadata",
              "software"
            ],
            "version":"2.0",
            "dateCreated":"2017-06-05",
            "datePublished":"2017-06-05",
            "programmingLanguage": "JSON-LD"
          }"""  # noqa
        expected_result = {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "type": "SoftwareSourceCode",
            "identifier": "CodeMeta",
            "description": "CodeMeta is a concept vocabulary that can "
            "be used to standardize the exchange of software metadata "
            "across repositories and organizations.",
            "name": "CodeMeta: Minimal metadata schemas for science "
            "software and code, in JSON-LD",
            "codeRepository": "https://github.com/codemeta/codemeta",
            "issueTracker": "https://github.com/codemeta/codemeta/issues",
            "license": "https://spdx.org/licenses/Apache-2.0",
            "version": "2.0",
            "author": [
                {
                    "type": "Person",
                    "givenName": "Carl",
                    "familyName": "Boettiger",
                    "email": "cboettig@gmail.com",
                    "id": "http://orcid.org/0000-0002-1642-628X",
                },
                {
                    "type": "Person",
                    "givenName": "Matthew B.",
                    "familyName": "Jones",
                    "email": "jones@nceas.ucsb.edu",
                    "id": "http://orcid.org/0000-0003-0077-4738",
                },
            ],
            "maintainer": {
                "type": "Person",
                "givenName": "Carl",
                "familyName": "Boettiger",
                "email": "cboettig@gmail.com",
                "id": "http://orcid.org/0000-0002-1642-628X",
            },
            "contIntegration": "https://travis-ci.org/codemeta/codemeta",
            "developmentStatus": "active",
            "downloadUrl": "https://github.com/codemeta/codemeta/archive/2.0.zip",
            "funder": {
                "id": "https://doi.org/10.13039/100000001",
                "type": "Organization",
                "name": "National Science Foundation",
            },
            "funding": "1549758; Codemeta: A Rosetta Stone for Metadata "
            "in Scientific Software",
            "keywords": ["metadata", "software"],
            "version": "2.0",
            "dateCreated": "2017-06-05",
            "datePublished": "2017-06-05",
            "programmingLanguage": "JSON-LD",
        }
        result = self.codemeta_mapping.translate(raw_content)
        self.assertEqual(result, expected_result)

    def test_compute_metadata_codemeta_alternate_context(self):
        raw_content = b"""{
            "@context": "https://raw.githubusercontent.com/codemeta/codemeta/master/codemeta.jsonld",
            "@type": "SoftwareSourceCode",
            "identifier": "CodeMeta"
        }"""  # noqa
        expected_result = {
            "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
            "type": "SoftwareSourceCode",
            "identifier": "CodeMeta",
        }
        result = self.codemeta_mapping.translate(raw_content)
        self.assertEqual(result, expected_result)

    def test_compute_metadata_maven(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <repositories>
            <repository>
              <id>central</id>
              <name>Maven Repository Switchboard</name>
              <layout>default</layout>
              <url>http://repo1.maven.org/maven2</url>
              <snapshots>
                <enabled>false</enabled>
              </snapshots>
            </repository>
          </repositories>
          <licenses>
            <license>
              <name>Apache License, Version 2.0</name>
              <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
              <distribution>repo</distribution>
              <comments>A business-friendly OSS license</comments>
            </license>
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "Maven Default Project",
                "identifier": "com.mycompany.app",
                "version": "1.2.3",
                "license": "https://www.apache.org/licenses/LICENSE-2.0.txt",
                "codeRepository": (
                    "http://repo1.maven.org/maven2/com/mycompany/app/my-app"
                ),
            },
        )

    def test_compute_metadata_maven_empty(self):
        raw_content = b"""
        <project>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
            },
        )

    def test_compute_metadata_maven_almost_empty(self):
        raw_content = b"""
        <project>
          <foo/>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
            },
        )

    def test_compute_metadata_maven_invalid_xml(self):
        expected_warning = (
            "WARNING:swh.indexer.metadata_dictionary.maven.MavenMapping:"
            "Error parsing XML from foo"
        )

        raw_content = b"""
        <project>"""
        with self.assertLogs("swh.indexer.metadata_dictionary", level="WARNING") as cm:
            result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

        raw_content = b"""
        """
        with self.assertLogs("swh.indexer.metadata_dictionary", level="WARNING") as cm:
            result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

    def test_compute_metadata_maven_unknown_encoding(self):
        expected_warning = (
            "WARNING:swh.indexer.metadata_dictionary.maven.MavenMapping:"
            "Error detecting XML encoding from foo"
        )

        raw_content = b"""<?xml version="1.0" encoding="foo"?>
        <project>
        </project>"""
        with self.assertLogs("swh.indexer.metadata_dictionary", level="WARNING") as cm:
            result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

        raw_content = b"""<?xml version="1.0" encoding="UTF-7"?>
        <project>
        </project>"""
        with self.assertLogs("swh.indexer.metadata_dictionary", level="WARNING") as cm:
            result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

    def test_compute_metadata_maven_invalid_encoding(self):
        expected_warning = (
            "WARNING:swh.indexer.metadata_dictionary.maven.MavenMapping:"
            "Error unidecoding XML from foo"
        )

        raw_content = b"""<?xml version="1.0" encoding="UTF-8"?>
        <foo\xe5ct>
        </foo>"""
        with self.assertLogs("swh.indexer.metadata_dictionary", level="WARNING") as cm:
            result = MAPPINGS["MavenMapping"]("foo").translate(raw_content)
            self.assertEqual(cm.output, [expected_warning])
        self.assertEqual(result, None)

    def test_compute_metadata_maven_minimal(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "Maven Default Project",
                "identifier": "com.mycompany.app",
                "version": "1.2.3",
                "codeRepository": (
                    "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
                ),
            },
        )

    def test_compute_metadata_maven_empty_nodes(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <repositories>
          </repositories>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "Maven Default Project",
                "identifier": "com.mycompany.app",
                "version": "1.2.3",
                "codeRepository": (
                    "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
                ),
            },
        )

        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version></version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "Maven Default Project",
                "identifier": "com.mycompany.app",
                "codeRepository": (
                    "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
                ),
            },
        )

        raw_content = b"""
        <project>
          <name></name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "identifier": "com.mycompany.app",
                "version": "1.2.3",
                "codeRepository": (
                    "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
                ),
            },
        )

        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <licenses>
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "Maven Default Project",
                "identifier": "com.mycompany.app",
                "version": "1.2.3",
                "codeRepository": (
                    "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
                ),
            },
        )

        raw_content = b"""
        <project>
          <groupId></groupId>
          <version>1.2.3</version>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "version": "1.2.3",
            },
        )

    def test_compute_metadata_maven_invalid_licenses(self):
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <licenses>
            foo
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "Maven Default Project",
                "identifier": "com.mycompany.app",
                "version": "1.2.3",
                "codeRepository": (
                    "https://repo.maven.apache.org/maven2/com/mycompany/app/my-app"
                ),
            },
        )

    def test_compute_metadata_maven_multiple(self):
        """Tests when there are multiple code repos and licenses."""
        raw_content = b"""
        <project>
          <name>Maven Default Project</name>
          <modelVersion>4.0.0</modelVersion>
          <groupId>com.mycompany.app</groupId>
          <artifactId>my-app</artifactId>
          <version>1.2.3</version>
          <repositories>
            <repository>
              <id>central</id>
              <name>Maven Repository Switchboard</name>
              <layout>default</layout>
              <url>http://repo1.maven.org/maven2</url>
              <snapshots>
                <enabled>false</enabled>
              </snapshots>
            </repository>
            <repository>
              <id>example</id>
              <name>Example Maven Repo</name>
              <layout>default</layout>
              <url>http://example.org/maven2</url>
            </repository>
          </repositories>
          <licenses>
            <license>
              <name>Apache License, Version 2.0</name>
              <url>https://www.apache.org/licenses/LICENSE-2.0.txt</url>
              <distribution>repo</distribution>
              <comments>A business-friendly OSS license</comments>
            </license>
            <license>
              <name>MIT license</name>
              <url>https://opensource.org/licenses/MIT</url>
            </license>
          </licenses>
        </project>"""
        result = self.maven_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "Maven Default Project",
                "identifier": "com.mycompany.app",
                "version": "1.2.3",
                "license": [
                    "https://www.apache.org/licenses/LICENSE-2.0.txt",
                    "https://opensource.org/licenses/MIT",
                ],
                "codeRepository": [
                    "http://repo1.maven.org/maven2/com/mycompany/app/my-app",
                    "http://example.org/maven2/com/mycompany/app/my-app",
                ],
            },
        )

    def test_compute_metadata_pkginfo(self):
        raw_content = b"""\
Metadata-Version: 2.1
Name: swh.core
Version: 0.0.49
Summary: Software Heritage core utilities
Home-page: https://forge.softwareheritage.org/diffusion/DCORE/
Author: Software Heritage developers
Author-email: swh-devel@inria.fr
License: UNKNOWN
Project-URL: Bug Reports, https://forge.softwareheritage.org/maniphest
Project-URL: Funding, https://www.softwareheritage.org/donate
Project-URL: Source, https://forge.softwareheritage.org/source/swh-core
Description: swh-core
        ========
       \x20
        core library for swh's modules:
        - config parser
        - hash computations
        - serialization
        - logging mechanism
       \x20
Platform: UNKNOWN
Classifier: Programming Language :: Python :: 3
Classifier: Intended Audience :: Developers
Classifier: License :: OSI Approved :: GNU General Public License v3 (GPLv3)
Classifier: Operating System :: OS Independent
Classifier: Development Status :: 5 - Production/Stable
Description-Content-Type: text/markdown
Provides-Extra: testing
"""  # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertCountEqual(
            result["description"],
            [
                "Software Heritage core utilities",  # note the comma here
                "swh-core\n"
                "========\n"
                "\n"
                "core library for swh's modules:\n"
                "- config parser\n"
                "- hash computations\n"
                "- serialization\n"
                "- logging mechanism\n"
                "",
            ],
            result,
        )
        del result["description"]
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "url": "https://forge.softwareheritage.org/diffusion/DCORE/",
                "name": "swh.core",
                "author": [
                    {
                        "type": "Person",
                        "name": "Software Heritage developers",
                        "email": "swh-devel@inria.fr",
                    }
                ],
                "version": "0.0.49",
            },
        )

    def test_compute_metadata_pkginfo_utf8(self):
        raw_content = b"""\
Metadata-Version: 1.1
Name: snowpyt
Description-Content-Type: UNKNOWN
Description: foo
        Hydrology N\xc2\xb083
"""  # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "snowpyt",
                "description": "foo\nHydrology N°83",
            },
        )

    def test_compute_metadata_pkginfo_keywords(self):
        raw_content = b"""\
Metadata-Version: 2.1
Name: foo
Keywords: foo bar baz
"""  # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "foo",
                "keywords": ["foo", "bar", "baz"],
            },
        )

    def test_compute_metadata_pkginfo_license(self):
        raw_content = b"""\
Metadata-Version: 2.1
Name: foo
License: MIT
"""  # noqa
        result = self.pkginfo_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "foo",
                "license": "MIT",
            },
        )

    def test_gemspec_base(self):
        raw_content = b"""
Gem::Specification.new do |s|
  s.name        = 'example'
  s.version     = '0.1.0'
  s.licenses    = ['MIT']
  s.summary     = "This is an example!"
  s.description = "Much longer explanation of the example!"
  s.authors     = ["Ruby Coder"]
  s.email       = 'rubycoder@example.com'
  s.files       = ["lib/example.rb"]
  s.homepage    = 'https://rubygems.org/gems/example'
  s.metadata    = { "source_code_uri" => "https://github.com/example/example" }
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertCountEqual(
            result.pop("description"),
            ["This is an example!", "Much longer explanation of the example!"],
        )
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "author": [{"type": "Person", "name": "Ruby Coder"}],
                "name": "example",
                "license": "https://spdx.org/licenses/MIT",
                "codeRepository": "https://rubygems.org/gems/example",
                "email": "rubycoder@example.com",
                "version": "0.1.0",
            },
        )

    def test_gemspec_two_author_fields(self):
        raw_content = b"""
Gem::Specification.new do |s|
  s.authors     = ["Ruby Coder1"]
  s.author      = "Ruby Coder2"
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertCountEqual(
            result.pop("author"),
            [
                {"type": "Person", "name": "Ruby Coder1"},
                {"type": "Person", "name": "Ruby Coder2"},
            ],
        )
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
            },
        )

    def test_gemspec_invalid_author(self):
        raw_content = b"""
Gem::Specification.new do |s|
  s.author      = ["Ruby Coder"]
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
            },
        )
        raw_content = b"""
Gem::Specification.new do |s|
  s.author      = "Ruby Coder1",
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
            },
        )
        raw_content = b"""
Gem::Specification.new do |s|
  s.authors     = ["Ruby Coder1", ["Ruby Coder2"]]
end"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "author": [{"type": "Person", "name": "Ruby Coder1"}],
            },
        )

    def test_gemspec_alternative_header(self):
        raw_content = b"""
require './lib/version'

Gem::Specification.new { |s|
  s.name = 'rb-system-with-aliases'
  s.summary = 'execute system commands with aliases'
}
"""
        result = self.gemspec_mapping.translate(raw_content)
        self.assertEqual(
            result,
            {
                "@context": "https://doi.org/10.5063/schema/codemeta-2.0",
                "type": "SoftwareSourceCode",
                "name": "rb-system-with-aliases",
                "description": "execute system commands with aliases",
            },
        )

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(json_document_strategy(keys=list(NpmMapping.mapping)))
    def test_npm_adversarial(self, doc):
        raw = json.dumps(doc).encode()
        self.npm_mapping.translate(raw)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(json_document_strategy(keys=CODEMETA_TERMS))
    def test_codemeta_adversarial(self, doc):
        raw = json.dumps(doc).encode()
        self.codemeta_mapping.translate(raw)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        xml_document_strategy(
            keys=list(MavenMapping.mapping),
            root="project",
            xmlns="http://maven.apache.org/POM/4.0.0",
        )
    )
    def test_maven_adversarial(self, doc):
        self.maven_mapping.translate(doc)

    @settings(suppress_health_check=[HealthCheck.too_slow])
    @given(
        strategies.dictionaries(
            # keys
            strategies.one_of(
                strategies.text(), *map(strategies.just, GemspecMapping.mapping)
            ),
            # values
            strategies.recursive(
                strategies.characters(),
                lambda children: strategies.lists(children, min_size=1),
            ),
        )
    )
    def test_gemspec_adversarial(self, doc):
        parts = [b"Gem::Specification.new do |s|\n"]
        for (k, v) in doc.items():
            parts.append("  s.{} = {}\n".format(k, repr(v)).encode())
        parts.append(b"end\n")
        self.gemspec_mapping.translate(b"".join(parts))

    def test_revision_metadata_indexer(self):
        metadata_indexer = RevisionMetadataIndexer(config=REVISION_METADATA_CONFIG)
        fill_obj_storage(metadata_indexer.objstorage)
        fill_storage(metadata_indexer.storage)

        tool = metadata_indexer.idx_storage.indexer_configuration_get(
            {f"tool_{k}": v for (k, v) in TRANSLATOR_TOOL.items()}
        )
        assert tool is not None
        rev = REVISION
        assert rev.directory == DIRECTORY2.id

        metadata_indexer.idx_storage.content_metadata_add(
            [
                ContentMetadataRow(
                    id=DIRECTORY2.entries[0].target,
                    indexer_configuration_id=tool["id"],
                    metadata=YARN_PARSER_METADATA,
                )
            ]
        )

        metadata_indexer.run([rev.id], "update-dups")

        results = list(
            metadata_indexer.idx_storage.revision_intrinsic_metadata_get([REVISION.id])
        )

        expected_results = [
            RevisionIntrinsicMetadataRow(
                id=rev.id,
                tool=TRANSLATOR_TOOL,
                metadata=YARN_PARSER_METADATA,
                mappings=["npm"],
            )
        ]

        for result in results:
            del result.tool["id"]

        # then
        self.assertEqual(results, expected_results)

    def test_revision_metadata_indexer_single_root_dir(self):
        metadata_indexer = RevisionMetadataIndexer(config=REVISION_METADATA_CONFIG)
        fill_obj_storage(metadata_indexer.objstorage)
        fill_storage(metadata_indexer.storage)

        # Add a parent directory, that is the only directory at the root
        # of the revision
        rev = REVISION
        assert rev.directory == DIRECTORY2.id

        directory = Directory(
            entries=(
                DirectoryEntry(
                    name=b"foobar-1.0.0", type="dir", target=rev.directory, perms=16384,
                ),
            ),
        )
        assert directory.id is not None
        metadata_indexer.storage.directory_add([directory])

        new_rev_dict = {**rev.to_dict(), "directory": directory.id}
        new_rev_dict.pop("id")
        new_rev = Revision.from_dict(new_rev_dict)
        metadata_indexer.storage.revision_add([new_rev])

        tool = metadata_indexer.idx_storage.indexer_configuration_get(
            {f"tool_{k}": v for (k, v) in TRANSLATOR_TOOL.items()}
        )
        assert tool is not None

        metadata_indexer.idx_storage.content_metadata_add(
            [
                ContentMetadataRow(
                    id=DIRECTORY2.entries[0].target,
                    indexer_configuration_id=tool["id"],
                    metadata=YARN_PARSER_METADATA,
                )
            ]
        )

        metadata_indexer.run([new_rev.id], "update-dups")

        results = list(
            metadata_indexer.idx_storage.revision_intrinsic_metadata_get([new_rev.id])
        )

        expected_results = [
            RevisionIntrinsicMetadataRow(
                id=new_rev.id,
                tool=TRANSLATOR_TOOL,
                metadata=YARN_PARSER_METADATA,
                mappings=["npm"],
            )
        ]

        for result in results:
            del result.tool["id"]

        # then
        self.assertEqual(results, expected_results)



# Copyright (C) 2015-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pytest

from swh.indexer.storage import get_indexer_storage
from swh.indexer.storage.api.client import RemoteStorage
import swh.indexer.storage.api.server as server

from .test_storage import *  # noqa


@pytest.fixture
def app(swh_indexer_storage_postgresql):
    server.storage = get_indexer_storage("local", db=swh_indexer_storage_postgresql.dsn)
    return server.app


@pytest.fixture
def swh_rpc_client_class():
    # these are needed for the swh_indexer_storage_with_data fixture
    assert hasattr(RemoteStorage, "indexer_configuration_add")
    assert hasattr(RemoteStorage, "content_mimetype_add")
    return RemoteStorage


@pytest.fixture
def swh_indexer_storage(swh_rpc_client, app):
    # This version of the swh_storage fixture uses the swh_rpc_client fixture
    # to instantiate a RemoteStorage (see swh_rpc_client_class above) that
    # proxies, via the swh.core RPC mechanism, the local (in memory) storage
    # configured in the app fixture above.
    return swh_rpc_client



# Copyright (C) 2016-2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import abc
from contextlib import contextmanager
import logging
import os
import shutil
import tempfile
from typing import Any, Dict, Generic, Iterator, List, Optional, Set, TypeVar, Union

from swh.core import utils
from swh.core.config import load_from_envvar, merge_configs
from swh.indexer.storage import INDEXER_CFG_KEY, PagedResult, Sha1, get_indexer_storage
from swh.indexer.storage.interface import IndexerStorageInterface
from swh.model import hashutil
from swh.model.model import Revision, Sha1Git
from swh.objstorage.exc import ObjNotFoundError
from swh.objstorage.factory import get_objstorage
from swh.scheduler import CONFIG as SWH_CONFIG
from swh.storage import get_storage
from swh.storage.interface import StorageInterface


@contextmanager
def write_to_temp(filename: str, data: bytes, working_directory: str) -> Iterator[str]:
    """Write the sha1's content in a temporary file.

    Args:
        filename: one of sha1's many filenames
        data: the sha1's content to write in temporary
          file
        working_directory: the directory into which the
          file is written

    Returns:
        The path to the temporary file created. That file is
        filled in with the raw content's data.

    """
    os.makedirs(working_directory, exist_ok=True)
    temp_dir = tempfile.mkdtemp(dir=working_directory)
    content_path = os.path.join(temp_dir, filename)

    with open(content_path, "wb") as f:
        f.write(data)

    yield content_path
    shutil.rmtree(temp_dir)


DEFAULT_CONFIG = {
    INDEXER_CFG_KEY: {"cls": "memory"},
    "storage": {"cls": "memory"},
    "objstorage": {"cls": "memory"},
}


TId = TypeVar("TId")
"""type of the ids of index()ed objects."""
TData = TypeVar("TData")
"""type of the objects passed to index()."""
TResult = TypeVar("TResult")
"""return type of index()"""


class BaseIndexer(Generic[TId, TData, TResult], metaclass=abc.ABCMeta):
    """Base class for indexers to inherit from.

    The main entry point is the :func:`run` function which is in
    charge of triggering the computations on the batch dict/ids
    received.

    Indexers can:

    - filter out ids whose data has already been indexed.
    - retrieve ids data from storage or objstorage
    - index this data depending on the object and store the result in
      storage.

    To implement a new object type indexer, inherit from the
    BaseIndexer and implement indexing:

    :meth:`~BaseIndexer.run`:
      object_ids are different depending on object. For example: sha1 for
      content, sha1_git for revision, directory, release, and id for origin

    To implement a new concrete indexer, inherit from the object level
    classes: :class:`ContentIndexer`, :class:`RevisionIndexer`,
    :class:`OriginIndexer`.

    Then you need to implement the following functions:

    :meth:`~BaseIndexer.filter`:
      filter out data already indexed (in storage).

    :meth:`~BaseIndexer.index_object`:
      compute index on id with data (retrieved from the storage or the
      objstorage by the id key) and return the resulting index computation.

    :meth:`~BaseIndexer.persist_index_computations`:
      persist the results of multiple index computations in the storage.

    The new indexer implementation can also override the following functions:

    :meth:`~BaseIndexer.prepare`:
      Configuration preparation for the indexer.  When overriding, this must
      call the `super().prepare()` instruction.

    :meth:`~BaseIndexer.check`:
      Configuration check for the indexer.  When overriding, this must call the
      `super().check()` instruction.

    :meth:`~BaseIndexer.register_tools`:
      This should return a dict of the tool(s) to use when indexing or
      filtering.

    """

    results: List[TResult]

    USE_TOOLS = True

    catch_exceptions = True
    """Prevents exceptions in `index()` from raising too high. Set to False
    in tests to properly catch all exceptions."""

    scheduler: Any
    storage: StorageInterface
    objstorage: Any
    idx_storage: IndexerStorageInterface

    def __init__(self, config=None, **kw) -> None:
        """Prepare and check that the indexer is ready to run.

        """
        super().__init__()
        if config is not None:
            self.config = config
        elif SWH_CONFIG:
            self.config = SWH_CONFIG.copy()
        else:
            self.config = load_from_envvar()
        self.config = merge_configs(DEFAULT_CONFIG, self.config)
        self.prepare()
        self.check()
        self.log.debug("%s: config=%s", self, self.config)

    def prepare(self) -> None:
        """Prepare the indexer's needed runtime configuration.
           Without this step, the indexer cannot possibly run.

        """
        config_storage = self.config.get("storage")
        if config_storage:
            self.storage = get_storage(**config_storage)

        objstorage = self.config["objstorage"]
        self.objstorage = get_objstorage(objstorage["cls"], objstorage["args"])

        idx_storage = self.config[INDEXER_CFG_KEY]
        self.idx_storage = get_indexer_storage(**idx_storage)

        _log = logging.getLogger("requests.packages.urllib3.connectionpool")
        _log.setLevel(logging.WARN)
        self.log = logging.getLogger("swh.indexer")

        if self.USE_TOOLS:
            self.tools = list(self.register_tools(self.config.get("tools", [])))
        self.results = []

    @property
    def tool(self) -> Dict:
        return self.tools[0]

    def check(self) -> None:
        """Check the indexer's configuration is ok before proceeding.
           If ok, does nothing. If not raise error.

        """
        if self.USE_TOOLS and not self.tools:
            raise ValueError("Tools %s is unknown, cannot continue" % self.tools)

    def _prepare_tool(self, tool: Dict[str, Any]) -> Dict[str, Any]:
        """Prepare the tool dict to be compliant with the storage api.

        """
        return {"tool_%s" % key: value for key, value in tool.items()}

    def register_tools(
        self, tools: Union[Dict[str, Any], List[Dict[str, Any]]]
    ) -> List[Dict[str, Any]]:
        """Permit to register tools to the storage.

           Add a sensible default which can be overridden if not
           sufficient.  (For now, all indexers use only one tool)

           Expects the self.config['tools'] property to be set with
           one or more tools.

        Args:
            tools: Either a dict or a list of dict.

        Returns:
            list: List of dicts with additional id key.

        Raises:
            ValueError: if not a list nor a dict.

        """
        if isinstance(tools, list):
            tools = list(map(self._prepare_tool, tools))
        elif isinstance(tools, dict):
            tools = [self._prepare_tool(tools)]
        else:
            raise ValueError("Configuration tool(s) must be a dict or list!")

        if tools:
            return self.idx_storage.indexer_configuration_add(tools)
        else:
            return []

    def index(self, id: TId, data: Optional[TData], **kwargs) -> List[TResult]:
        """Index computation for the id and associated raw data.

        Args:
            id: identifier or Dict object
            data: id's data from storage or objstorage depending on
              object type

        Returns:
            dict: a dict that makes sense for the
            :meth:`.persist_index_computations` method.

        """
        raise NotImplementedError()

    def filter(self, ids: List[TId]) -> Iterator[TId]:
        """Filter missing ids for that particular indexer.

        Args:
            ids: list of ids

        Yields:
            iterator of missing ids

        """
        yield from ids

    @abc.abstractmethod
    def persist_index_computations(
        self, results: List[TResult], policy_update: str
    ) -> Dict[str, int]:
        """Persist the computation resulting from the index.

        Args:

            results: List of results. One result is the
              result of the index function.
            policy_update: either 'update-dups' or 'ignore-dups' to
              respectively update duplicates or ignore them

        Returns:
            a summary dict of what has been inserted in the storage

        """
        return {}


class ContentIndexer(BaseIndexer[Sha1, bytes, TResult], Generic[TResult]):
    """A content indexer working on a list of ids directly.

    To work on indexer partition, use the :class:`ContentPartitionIndexer`
    instead.

    Note: :class:`ContentIndexer` is not an instantiable object. To
    use it, one should inherit from this class and override the
    methods mentioned in the :class:`BaseIndexer` class.

    """

    def run(self, ids: List[Sha1], policy_update: str, **kwargs) -> Dict:
        """Given a list of ids:

        - retrieve the content from the storage
        - execute the indexing computations
        - store the results (according to policy_update)

        Args:
            ids (Iterable[Union[bytes, str]]): sha1's identifier list
            policy_update (str): either 'update-dups' or 'ignore-dups' to
                                 respectively update duplicates or ignore
                                 them
            **kwargs: passed to the `index` method

        Returns:
            A summary Dict of the task's status

        """
        sha1s = [
            hashutil.hash_to_bytes(id_) if isinstance(id_, str) else id_ for id_ in ids
        ]
        results = []
        summary: Dict = {"status": "uneventful"}
        try:
            for sha1 in sha1s:
                try:
                    raw_content = self.objstorage.get(sha1)
                except ObjNotFoundError:
                    self.log.warning(
                        "Content %s not found in objstorage"
                        % hashutil.hash_to_hex(sha1)
                    )
                    continue
                res = self.index(sha1, raw_content, **kwargs)
                if res:  # If no results, skip it
                    results.extend(res)
                    summary["status"] = "eventful"
            summary = self.persist_index_computations(results, policy_update)
            self.results = results
        except Exception:
            if not self.catch_exceptions:
                raise
            self.log.exception("Problem when reading contents metadata.")
            summary["status"] = "failed"
        return summary


class ContentPartitionIndexer(BaseIndexer[Sha1, bytes, TResult], Generic[TResult]):
    """A content partition indexer.

    This expects as input a partition_id and a nb_partitions. This will then index the
    contents within that partition.

    To work on a list of ids, use the :class:`ContentIndexer` instead.

    Note: :class:`ContentPartitionIndexer` is not an instantiable
    object. To use it, one should inherit from this class and override
    the methods mentioned in the :class:`BaseIndexer` class.

    """

    @abc.abstractmethod
    def indexed_contents_in_partition(
        self, partition_id: int, nb_partitions: int, page_token: Optional[str] = None
    ) -> PagedResult[Sha1]:
        """Retrieve indexed contents within range [start, end].

        Args:
            partition_id: Index of the partition to fetch
            nb_partitions: Total number of partitions to split into
            page_token: opaque token used for pagination

        Returns:
            PagedResult of Sha1. If next_page_token is None, there is no more data
            to fetch

        """
        pass

    def _list_contents_to_index(
        self, partition_id: int, nb_partitions: int, indexed: Set[Sha1]
    ) -> Iterator[Sha1]:
        """Compute from storage the new contents to index in the partition_id . The already
           indexed contents are skipped.

        Args:
            partition_id: Index of the partition to fetch data from
            nb_partitions: Total number of partition
            indexed: Set of content already indexed.

        Yields:
            Sha1 id (bytes) of contents to index

        """
        if not isinstance(partition_id, int) or not isinstance(nb_partitions, int):
            raise TypeError(
                f"identifiers must be int, not {partition_id!r} and {nb_partitions!r}."
            )
        next_page_token = None
        while True:
            result = self.storage.content_get_partition(
                partition_id, nb_partitions, page_token=next_page_token
            )
            contents = result.results
            for c in contents:
                _id = hashutil.hash_to_bytes(c.sha1)
                if _id in indexed:
                    continue
                yield _id
            next_page_token = result.next_page_token
            if next_page_token is None:
                break

    def _index_contents(
        self, partition_id: int, nb_partitions: int, indexed: Set[Sha1], **kwargs: Any
    ) -> Iterator[TResult]:
        """Index the contents within the partition_id.

        Args:
            start: Starting bound from range identifier
            end: End range identifier
            indexed: Set of content already indexed.

        Yields:
            indexing result as dict to persist in the indexer backend

        """
        for sha1 in self._list_contents_to_index(partition_id, nb_partitions, indexed):
            try:
                raw_content = self.objstorage.get(sha1)
            except ObjNotFoundError:
                self.log.warning(f"Content {sha1.hex()} not found in objstorage")
                continue
            yield from self.index(sha1, raw_content, **kwargs)

    def _index_with_skipping_already_done(
        self, partition_id: int, nb_partitions: int
    ) -> Iterator[TResult]:
        """Index not already indexed contents within the partition partition_id

        Args:
            partition_id: Index of the partition to fetch
            nb_partitions: Total number of partitions to split into

        Yields:
           indexing result as dict to persist in the indexer backend

        """
        next_page_token = None
        contents = set()
        while True:
            indexed_page = self.indexed_contents_in_partition(
                partition_id, nb_partitions, page_token=next_page_token
            )
            for sha1 in indexed_page.results:
                contents.add(sha1)
            yield from self._index_contents(partition_id, nb_partitions, contents)
            next_page_token = indexed_page.next_page_token
            if next_page_token is None:
                break

    def run(
        self,
        partition_id: int,
        nb_partitions: int,
        skip_existing: bool = True,
        **kwargs,
    ) -> Dict:
        """Given a partition of content ids, index the contents within.

           Either the indexer is incremental (filter out existing computed data) or it
           computes everything from scratch.

        Args:
            partition_id: Index of the partition to fetch
            nb_partitions: Total number of partitions to split into
            skip_existing: Skip existing indexed data
                (default) or not
            **kwargs: passed to the `index` method

        Returns:
            dict with the indexing task status

        """
        summary: Dict[str, Any] = {"status": "uneventful"}
        count = 0
        try:
            if skip_existing:
                gen = self._index_with_skipping_already_done(
                    partition_id, nb_partitions
                )
            else:
                gen = self._index_contents(partition_id, nb_partitions, indexed=set([]))

            count_object_added_key: Optional[str] = None

            for contents in utils.grouper(gen, n=self.config["write_batch_size"]):
                res = self.persist_index_computations(
                    list(contents), policy_update="update-dups"
                )
                if not count_object_added_key:
                    count_object_added_key = list(res.keys())[0]
                count += res[count_object_added_key]
                if count > 0:
                    summary["status"] = "eventful"
        except Exception:
            if not self.catch_exceptions:
                raise
            self.log.exception("Problem when computing metadata.")
            summary["status"] = "failed"

        if count > 0 and count_object_added_key:
            summary[count_object_added_key] = count
        return summary


class OriginIndexer(BaseIndexer[str, None, TResult], Generic[TResult]):
    """An object type indexer, inherits from the :class:`BaseIndexer` and
    implements Origin indexing using the run method

    Note: the :class:`OriginIndexer` is not an instantiable object.
    To use it in another context one should inherit from this class
    and override the methods mentioned in the :class:`BaseIndexer`
    class.

    """

    def run(
        self, origin_urls: List[str], policy_update: str = "update-dups", **kwargs
    ) -> Dict:
        """Given a list of origin urls:

        - retrieve origins from storage
        - execute the indexing computations
        - store the results (according to policy_update)

        Args:
            origin_urls: list of origin urls.
            policy_update: either 'update-dups' or 'ignore-dups' to
              respectively update duplicates (default) or ignore them
            **kwargs: passed to the `index` method

        """
        summary: Dict[str, Any] = {"status": "uneventful"}
        try:
            results = self.index_list(origin_urls, **kwargs)
        except Exception:
            if not self.catch_exceptions:
                raise
            summary["status"] = "failed"
            return summary

        summary_persist = self.persist_index_computations(results, policy_update)
        self.results = results
        if summary_persist:
            for value in summary_persist.values():
                if value > 0:
                    summary["status"] = "eventful"
            summary.update(summary_persist)
        return summary

    def index_list(self, origin_urls: List[str], **kwargs) -> List[TResult]:
        results = []
        for origin_url in origin_urls:
            try:
                results.extend(self.index(origin_url, **kwargs))
            except Exception:
                self.log.exception("Problem when processing origin %s", origin_url)
                raise
        return results


class RevisionIndexer(BaseIndexer[Sha1Git, Revision, TResult], Generic[TResult]):
    """An object type indexer, inherits from the :class:`BaseIndexer` and
    implements Revision indexing using the run method

    Note: the :class:`RevisionIndexer` is not an instantiable object.
    To use it in another context one should inherit from this class
    and override the methods mentioned in the :class:`BaseIndexer`
    class.

    """

    def run(self, ids: List[Sha1Git], policy_update: str) -> Dict:
        """Given a list of sha1_gits:

        - retrieve revisions from storage
        - execute the indexing computations
        - store the results (according to policy_update)

        Args:
            ids: sha1_git's identifier list
            policy_update: either 'update-dups' or 'ignore-dups' to
              respectively update duplicates or ignore them

        """
        summary: Dict[str, Any] = {"status": "uneventful"}
        results = []

        revision_ids = [
            hashutil.hash_to_bytes(id_) if isinstance(id_, str) else id_ for id_ in ids
        ]
        for (rev_id, rev) in zip(revision_ids, self.storage.revision_get(revision_ids)):
            if not rev:
                # TODO: call self.index() with rev=None?
                self.log.warning(
                    "Revision %s not found in storage", hashutil.hash_to_hex(rev_id)
                )
                continue
            try:
                results.extend(self.index(rev_id, rev))
            except Exception:
                if not self.catch_exceptions:
                    raise
                self.log.exception("Problem when processing revision")
                summary["status"] = "failed"
                return summary

        summary_persist = self.persist_index_computations(results, policy_update)
        if summary_persist:
            for value in summary_persist.values():
                if value > 0:
                    summary["status"] = "eventful"
            summary.update(summary_persist)
        self.results = results
        return summary



from celery import current_app as app

from swh.indexer.metadata import OriginMetadataIndexer, RevisionMetadataIndexer

from .test_metadata import ContentMetadataTestIndexer
from .test_origin_head import OriginHeadTestIndexer
from .utils import BASE_TEST_CONFIG


class RevisionMetadataTestIndexer(RevisionMetadataIndexer):
    """Specific indexer whose configuration is enough to satisfy the
       indexing tests.
    """

    ContentMetadataIndexer = ContentMetadataTestIndexer

    def parse_config_file(self, *args, **kwargs):
        return {
            **BASE_TEST_CONFIG,
            "tools": {
                "name": "swh-metadata-detector",
                "version": "0.0.2",
                "configuration": {"type": "local", "context": "NpmMapping"},
            },
        }


class OriginMetadataTestIndexer(OriginMetadataIndexer):
    def parse_config_file(self, *args, **kwargs):
        return {**BASE_TEST_CONFIG, "tools": []}

    def _prepare_sub_indexers(self):
        self.origin_head_indexer = OriginHeadTestIndexer()
        self.revision_metadata_indexer = RevisionMetadataTestIndexer()


@app.task
def revision_intrinsic_metadata(*args, **kwargs):
    indexer = RevisionMetadataTestIndexer()
    indexer.run(*args, **kwargs)
    print("REV RESULT=", indexer.results)


@app.task
def origin_intrinsic_metadata(*args, **kwargs):
    indexer = OriginMetadataTestIndexer()
    indexer.run(*args, **kwargs)



# Copyright (C) 2019 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import unittest
from unittest.mock import Mock, patch

from swh.indexer.journal_client import process_journal_objects


class JournalClientTest(unittest.TestCase):
    def testOneOriginVisit(self):
        mock_scheduler = Mock()
        messages = {
            "origin_visit": [{"status": "full", "origin": "file:///dev/zero",},]
        }
        process_journal_objects(
            messages,
            scheduler=mock_scheduler,
            task_names={"origin_metadata": "task-name"},
        )
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]["next_run"]
        self.assertEqual(
            args,
            (
                [
                    {
                        "arguments": {
                            "kwargs": {"policy_update": "update-dups"},
                            "args": (["file:///dev/zero"],),
                        },
                        "policy": "oneshot",
                        "type": "task-name",
                    },
                ],
            ),
        )

    def testOriginVisitLegacy(self):
        mock_scheduler = Mock()
        messages = {
            "origin_visit": [
                {"status": "full", "origin": {"url": "file:///dev/zero",}},
            ]
        }
        process_journal_objects(
            messages,
            scheduler=mock_scheduler,
            task_names={"origin_metadata": "task-name"},
        )
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]["next_run"]
        self.assertEqual(
            args,
            (
                [
                    {
                        "arguments": {
                            "kwargs": {"policy_update": "update-dups"},
                            "args": (["file:///dev/zero"],),
                        },
                        "policy": "oneshot",
                        "type": "task-name",
                    },
                ],
            ),
        )

    def testOneOriginVisitBatch(self):
        mock_scheduler = Mock()
        messages = {
            "origin_visit": [
                {"status": "full", "origin": "file:///dev/zero",},
                {"status": "full", "origin": "file:///tmp/foobar",},
            ]
        }
        process_journal_objects(
            messages,
            scheduler=mock_scheduler,
            task_names={"origin_metadata": "task-name"},
        )
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]["next_run"]
        self.assertEqual(
            args,
            (
                [
                    {
                        "arguments": {
                            "kwargs": {"policy_update": "update-dups"},
                            "args": (["file:///dev/zero", "file:///tmp/foobar"],),
                        },
                        "policy": "oneshot",
                        "type": "task-name",
                    },
                ],
            ),
        )

    @patch("swh.indexer.journal_client.MAX_ORIGINS_PER_TASK", 2)
    def testOriginVisitBatches(self):
        mock_scheduler = Mock()
        messages = {
            "origin_visit": [
                {"status": "full", "origin": "file:///dev/zero",},
                {"status": "full", "origin": "file:///tmp/foobar",},
                {"status": "full", "origin": "file:///tmp/spamegg",},
            ]
        }
        process_journal_objects(
            messages,
            scheduler=mock_scheduler,
            task_names={"origin_metadata": "task-name"},
        )
        self.assertTrue(mock_scheduler.create_tasks.called)
        call_args = mock_scheduler.create_tasks.call_args
        (args, kwargs) = call_args
        self.assertEqual(kwargs, {})
        del args[0][0]["next_run"]
        del args[0][1]["next_run"]
        self.assertEqual(
            args,
            (
                [
                    {
                        "arguments": {
                            "kwargs": {"policy_update": "update-dups"},
                            "args": (["file:///dev/zero", "file:///tmp/foobar"],),
                        },
                        "policy": "oneshot",
                        "type": "task-name",
                    },
                    {
                        "arguments": {
                            "kwargs": {"policy_update": "update-dups"},
                            "args": (["file:///tmp/spamegg"],),
                        },
                        "policy": "oneshot",
                        "type": "task-name",
                    },
                ],
            ),
        )



# Copyright (C) 2020  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from typing import Any, Dict, List, Optional
from unittest.mock import Mock

import pytest

from swh.indexer.indexer import (
    ContentIndexer,
    ContentPartitionIndexer,
    OriginIndexer,
    RevisionIndexer,
)
from swh.indexer.storage import PagedResult, Sha1

from .utils import BASE_TEST_CONFIG


class _TestException(Exception):
    pass


class CrashingIndexerMixin:
    USE_TOOLS = False

    def index(
        self, id: Any, data: Optional[Any] = None, **kwargs
    ) -> List[Dict[str, Any]]:
        raise _TestException()

    def persist_index_computations(self, results, policy_update) -> Dict[str, int]:
        return {}

    def indexed_contents_in_partition(
        self, partition_id: int, nb_partitions: int, page_token: Optional[str] = None
    ) -> PagedResult[Sha1]:
        raise _TestException()


class CrashingContentIndexer(CrashingIndexerMixin, ContentIndexer):
    pass


class CrashingContentPartitionIndexer(CrashingIndexerMixin, ContentPartitionIndexer):
    pass


class CrashingRevisionIndexer(CrashingIndexerMixin, RevisionIndexer):
    pass


class CrashingOriginIndexer(CrashingIndexerMixin, OriginIndexer):
    pass


def test_content_indexer_catch_exceptions():
    indexer = CrashingContentIndexer(config=BASE_TEST_CONFIG)
    indexer.objstorage = Mock()
    indexer.objstorage.get.return_value = b"content"

    assert indexer.run([b"foo"], policy_update=True) == {"status": "failed"}

    indexer.catch_exceptions = False

    with pytest.raises(_TestException):
        indexer.run([b"foo"], policy_update=True)


def test_revision_indexer_catch_exceptions():
    indexer = CrashingRevisionIndexer(config=BASE_TEST_CONFIG)
    indexer.storage = Mock()
    indexer.storage.revision_get.return_value = ["rev"]

    assert indexer.run([b"foo"], policy_update=True) == {"status": "failed"}

    indexer.catch_exceptions = False

    with pytest.raises(_TestException):
        indexer.run([b"foo"], policy_update=True)


def test_origin_indexer_catch_exceptions():
    indexer = CrashingOriginIndexer(config=BASE_TEST_CONFIG)

    assert indexer.run(["http://example.org"], policy_update=True) == {
        "status": "failed"
    }

    indexer.catch_exceptions = False

    with pytest.raises(_TestException):
        indexer.run(["http://example.org"], policy_update=True)


def test_content_partition_indexer_catch_exceptions():
    indexer = CrashingContentPartitionIndexer(
        config={**BASE_TEST_CONFIG, "write_batch_size": 42}
    )

    assert indexer.run(0, 42, policy_update=True) == {"status": "failed"}

    indexer.catch_exceptions = False

    with pytest.raises(_TestException):
        indexer.run(0, 42, policy_update=True)



# Copyright (C) 2015-2017  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information


def ctags_to_db(ctags):
    """Convert a ctags entry into a ready ctags entry.

    Args:
        ctags (dict): ctags entry with the following keys:

            - id (bytes): content's identifier
            - tool_id (int): tool id used to compute ctags
            - ctags ([dict]): List of dictionary with the following keys:

              - name (str): symbol's name
              - kind (str): symbol's kind
              - line (int): symbol's line in the content
              - language (str): language

    Returns:
        list: list of ctags entries as dicts with the following keys:

        - id (bytes): content's identifier
        - name (str): symbol's name
        - kind (str): symbol's kind
        - language (str): language for that content
        - tool_id (int): tool id used to compute ctags

    """
    id = ctags["id"]
    tool_id = ctags["indexer_configuration_id"]
    for ctag in ctags["ctags"]:
        yield {
            "id": id,
            "name": ctag["name"],
            "kind": ctag["kind"],
            "line": ctag["line"],
            "lang": ctag["lang"],
            "indexer_configuration_id": tool_id,
        }


def db_to_ctags(ctag):
    """Convert a ctags entry into a ready ctags entry.

    Args:
        ctags (dict): ctags entry with the following keys:

          - id (bytes): content's identifier
          - ctags ([dict]): List of dictionary with the following keys:
            - name (str): symbol's name
            - kind (str): symbol's kind
            - line (int): symbol's line in the content
            - language (str): language

    Returns:
        list: list of ctags ready entry (dict with the following keys):

        - id (bytes): content's identifier
        - name (str): symbol's name
        - kind (str): symbol's kind
        - language (str): language for that content
        - tool (dict): tool used to compute the ctags

    """
    return {
        "id": ctag["id"],
        "name": ctag["name"],
        "kind": ctag["kind"],
        "line": ctag["line"],
        "lang": ctag["lang"],
        "tool": {
            "id": ctag["tool_id"],
            "name": ctag["tool_name"],
            "version": ctag["tool_version"],
            "configuration": ctag["tool_configuration"],
        },
    }


def db_to_mimetype(mimetype):
    """Convert a ctags entry into a ready ctags output.

    """
    return {
        "id": mimetype["id"],
        "encoding": mimetype["encoding"],
        "mimetype": mimetype["mimetype"],
        "tool": {
            "id": mimetype["tool_id"],
            "name": mimetype["tool_name"],
            "version": mimetype["tool_version"],
            "configuration": mimetype["tool_configuration"],
        },
    }


def db_to_language(language):
    """Convert a language entry into a ready language output.

    """
    return {
        "id": language["id"],
        "lang": language["lang"],
        "tool": {
            "id": language["tool_id"],
            "name": language["tool_name"],
            "version": language["tool_version"],
            "configuration": language["tool_configuration"],
        },
    }


def db_to_metadata(metadata):
    """Convert a metadata entry into a ready metadata output.

    """
    metadata["tool"] = {
        "id": metadata["tool_id"],
        "name": metadata["tool_name"],
        "version": metadata["tool_version"],
        "configuration": metadata["tool_configuration"],
    }
    del metadata["tool_id"], metadata["tool_configuration"]
    del metadata["tool_version"], metadata["tool_name"]
    return metadata


def db_to_fossology_license(license):
    return {
        "id": license["id"],
        "license": license["license"],
        "tool": {
            "id": license["tool_id"],
            "name": license["tool_name"],
            "version": license["tool_version"],
            "configuration": license["tool_configuration"],
        },
    }



# Copyright (C) 2018-2019  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import re

from swh.indexer.codemeta import CROSSWALK_TABLE, SCHEMA_URI

from .base import JsonMapping


class NpmMapping(JsonMapping):
    """
    dedicated class for NPM (package.json) mapping and translation
    """

    name = "npm"
    mapping = CROSSWALK_TABLE["NodeJS"]
    filename = b"package.json"
    string_fields = ["name", "version", "homepage", "description", "email"]

    _schema_shortcuts = {
        "github": "git+https://github.com/%s.git",
        "gist": "git+https://gist.github.com/%s.git",
        "gitlab": "git+https://gitlab.com/%s.git",
        # Bitbucket supports both hg and git, and the shortcut does not
        # tell which one to use.
        # 'bitbucket': 'https://bitbucket.org/',
    }

    def normalize_repository(self, d):
        """https://docs.npmjs.com/files/package.json#repository

        >>> NpmMapping().normalize_repository({
        ...     'type': 'git',
        ...     'url': 'https://example.org/foo.git'
        ... })
        {'@id': 'git+https://example.org/foo.git'}
        >>> NpmMapping().normalize_repository(
        ...     'gitlab:foo/bar')
        {'@id': 'git+https://gitlab.com/foo/bar.git'}
        >>> NpmMapping().normalize_repository(
        ...     'foo/bar')
        {'@id': 'git+https://github.com/foo/bar.git'}
        """
        if (
            isinstance(d, dict)
            and isinstance(d.get("type"), str)
            and isinstance(d.get("url"), str)
        ):
            url = "{type}+{url}".format(**d)
        elif isinstance(d, str):
            if "://" in d:
                url = d
            elif ":" in d:
                (schema, rest) = d.split(":", 1)
                if schema in self._schema_shortcuts:
                    url = self._schema_shortcuts[schema] % rest
                else:
                    return None
            else:
                url = self._schema_shortcuts["github"] % d

        else:
            return None

        return {"@id": url}

    def normalize_bugs(self, d):
        """https://docs.npmjs.com/files/package.json#bugs

        >>> NpmMapping().normalize_bugs({
        ...     'url': 'https://example.org/bugs/',
        ...     'email': 'bugs@example.org'
        ... })
        {'@id': 'https://example.org/bugs/'}
        >>> NpmMapping().normalize_bugs(
        ...     'https://example.org/bugs/')
        {'@id': 'https://example.org/bugs/'}
        """
        if isinstance(d, dict) and isinstance(d.get("url"), str):
            return {"@id": d["url"]}
        elif isinstance(d, str):
            return {"@id": d}
        else:
            return None

    _parse_author = re.compile(
        r"^ *" r"(?P<name>.*?)" r"( +<(?P<email>.*)>)?" r"( +\((?P<url>.*)\))?" r" *$"
    )

    def normalize_author(self, d):
        """https://docs.npmjs.com/files/package.json#people-fields-author-contributors'

        >>> from pprint import pprint
        >>> pprint(NpmMapping().normalize_author({
        ...     'name': 'John Doe',
        ...     'email': 'john.doe@example.org',
        ...     'url': 'https://example.org/~john.doe',
        ... }))
        {'@list': [{'@type': 'http://schema.org/Person',
                    'http://schema.org/email': 'john.doe@example.org',
                    'http://schema.org/name': 'John Doe',
                    'http://schema.org/url': {'@id': 'https://example.org/~john.doe'}}]}
        >>> pprint(NpmMapping().normalize_author(
        ...     'John Doe <john.doe@example.org> (https://example.org/~john.doe)'
        ... ))
        {'@list': [{'@type': 'http://schema.org/Person',
                    'http://schema.org/email': 'john.doe@example.org',
                    'http://schema.org/name': 'John Doe',
                    'http://schema.org/url': {'@id': 'https://example.org/~john.doe'}}]}
        """  # noqa
        author = {"@type": SCHEMA_URI + "Person"}
        if isinstance(d, dict):
            name = d.get("name", None)
            email = d.get("email", None)
            url = d.get("url", None)
        elif isinstance(d, str):
            match = self._parse_author.match(d)
            if not match:
                return None
            name = match.group("name")
            email = match.group("email")
            url = match.group("url")
        else:
            return None
        if name and isinstance(name, str):
            author[SCHEMA_URI + "name"] = name
        if email and isinstance(email, str):
            author[SCHEMA_URI + "email"] = email
        if url and isinstance(url, str):
            author[SCHEMA_URI + "url"] = {"@id": url}
        return {"@list": [author]}

    def normalize_license(self, s):
        """https://docs.npmjs.com/files/package.json#license

        >>> NpmMapping().normalize_license('MIT')
        {'@id': 'https://spdx.org/licenses/MIT'}
        """
        if isinstance(s, str):
            return {"@id": "https://spdx.org/licenses/" + s}

    def normalize_homepage(self, s):
        """https://docs.npmjs.com/files/package.json#homepage

        >>> NpmMapping().normalize_homepage('https://example.org/~john.doe')
        {'@id': 'https://example.org/~john.doe'}
        """
        if isinstance(s, str):
            return {"@id": s}

    def normalize_keywords(self, lst):
        """https://docs.npmjs.com/files/package.json#homepage

        >>> NpmMapping().normalize_keywords(['foo', 'bar'])
        ['foo', 'bar']
        """
        if isinstance(lst, list):
            return [x for x in lst if isinstance(x, str)]
